# SPDX-FileCopyrightText: 2020 UMG MeDIC <marcel.parciak@med.uni-goettingen.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import multiprocessing
import os
import shutil
import tempfile
import xml.etree.ElementTree as ET
import zipfile
from typing import Iterator, Tuple

from sqlalchemy.orm import Session

from tos_execute import config
from tos_execute import database
from tos_execute import error_log
from tos_execute import models_internal as intmodels
from tos_execute import queue
from tos_execute import store


def is_job_directory(path: str) -> bool:
    """Check for jobInfo.properties in order to assume the given path is a TOS job directory."""

    for entry in os.scandir(path):
        if entry.is_file() and entry.name == "jobInfo.properties":
            return True
    return False


def is_job_zipfile(path: str) -> bool:
    """Check for jobInfo.properties in order to assume the given path is a TOS job zipfile."""

    with zipfile.ZipFile(path) as zip:
        for member in zip.namelist():
            if member[-len("_run.sh") :] == "_run.sh":
                return True
    return False


def retrieve_job_executables(path: str) -> Iterator[str]:
    """Read the jobInfo.properties and extract the jobName from it."""

    for root, dirs, files in os.walk(path):
        for file in files:
            if file[-len("_run.sh") :] == "_run.sh":
                yield os.path.join(root, file)


def find_tos_execs(path: str) -> str:
    """Retrieves the first .sh script under path. Usually, this is used to find the shell execution script of a TOS job."""

    for root, dirs, files in os.walk(path):
        for file in files:
            if file[-3:] == ".sh":
                return os.path.join(root, file)
    return ""


def copy_directory_for_executable(
    executable: str, store_dir: str, job_identifier: Tuple[str, str]
) -> str:
    """Copies all necessary directories and files to `store_dir` needed to execute `executable`, returns written target directory path."""
    job_root_dir = os.path.dirname(os.path.abspath(executable))
    job_lib_dir = os.path.join(job_root_dir, os.pardir, "lib")
    job_properties_file = os.path.join(job_root_dir, os.pardir, "jobInfo.properties")
    job_name = "_".join(job_identifier)
    shutil.copytree(
        job_root_dir,
        os.path.join(store_dir, job_name, job_identifier[0]),
        dirs_exist_ok=True,
    )
    shutil.copytree(
        job_lib_dir, os.path.join(store_dir, job_name, "lib"), dirs_exist_ok=True
    )
    shutil.copyfile(
        job_properties_file, os.path.join(store_dir, job_name, "jobInfo.properties")
    )
    return find_tos_execs(os.path.join(store_dir, job_name))


def retrieve_job_name_and_version(executable: str) -> Tuple[str, str]:
    """Retrieves the job name and version based on the path of a executable of a Job. Will use the .properties generated by Talend file for this."""
    job_root_dir = os.path.dirname(os.path.abspath(executable))
    executable_filename = os.path.basename(executable)
    properties_file = None
    for root, dirs, files in os.walk(os.path.join(job_root_dir, "items")):
        for file in files:
            if file.startswith(
                executable_filename[: -len("_run.sh")]
            ) and file.endswith(".properties"):
                properties_file = os.path.join(root, file)

    if properties_file:
        tree = ET.parse(properties_file)
        for prop in tree.findall("{http://www.talend.org/properties}Property"):
            if prop.get("displayName") is not None and prop.get("version") is not None:
                return str(prop.get("displayName")), str(prop.get("version"))
        return ("", "")
    else:
        return ("", "")


def init_job_map(settings: config.BasicSettings) -> None:
    """Initializes all available TOS jobs from the configured base path."""

    base_dir = settings.talend_base_folder
    store_dir = settings.talend_job_cache
    joblist = []
    for entry in os.scandir(base_dir):
        # handle each zip file
        if entry.is_file() and entry.name[-4:] == ".zip" and is_job_zipfile(entry.path):
            with tempfile.TemporaryDirectory() as tempdir:
                with zipfile.ZipFile(entry.path, "r") as zip:
                    zip.extractall(tempdir)
                    for executable in retrieve_job_executables(tempdir):
                        name, version = retrieve_job_name_and_version(executable)
                        joblist.append(
                            {
                                "name": name,
                                "version": version,
                                "path": copy_directory_for_executable(
                                    executable,
                                    store_dir,
                                    job_identifier=(name, version),
                                ),
                            }
                        )

        # handle each directory
        if entry.is_dir() and is_job_directory(entry.path):
            for executable in retrieve_job_executables(entry.path):
                name, version = retrieve_job_name_and_version(executable)
                joblist.append(
                    {
                        "name": name,
                        "version": version,
                        "path": copy_directory_for_executable(
                            executable,
                            store_dir,
                            job_identifier=(name, version),
                        ),
                    }
                )

    db = database.SessionLocal()
    store.set_all_jobs_inactive(db=db)
    for job in joblist:
        job_entry = store.get_job(name=job["name"], version=job["version"], db=db)
        if not job_entry:
            job_entry = intmodels.Job(**job)
        job_entry.path = job["path"]
        job_entry.available = True
        store.store_job(job_entry, db=db)


def restart_queued_jobs():
    db: Session = database.SessionLocal()
    job_runs = store.get_job_runs_by_state(intmodels.JobState.queued, db=db)

    for job_run in job_runs:
        p = multiprocessing.Process(
            target=queue.start_job_process, args=[job_run.id], daemon=True
        )
        p.start()
