# SPDX-FileCopyrightText: 2020 UMG MeDIC <marcel.parciak@med.uni-goettingen.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import abc
import datetime
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field, validator

from tos_execute import config

#
# Common models
#
class CredentialsCommon(BaseModel):
    """
    This model represents the expected credentials content to communicate credentials with the agent.
    """

    name: str = Field("", example="")
    value: str = Field("", example="")

    @staticmethod
    def example() -> dict:
        return {"name": "", "value": ""}


class ParamsCommon(BaseModel, abc.ABC):
    """
    These are all possible parameters that may be supplied using the Active Workflow Remote Agent API.
    """

    message: Optional[Dict[str, Any]] = {}
    options: Optional[Dict[str, Any]] = {}
    memory: Optional[Dict[str, Any]] = {}
    credentials: Optional[List[CredentialsCommon]] = []


class ResultCommon(BaseModel, abc.ABC):
    """
    These are all possible result values that may be returned using the Active Workflow Remote Agent API.
    """

    errors: Optional[List[str]] = None
    logs: Optional[List[str]] = None
    memory: Optional[Dict[str, Any]] = None
    messages: Optional[List[Dict[str, Any]]] = None
    name: Optional[str] = None
    display_name: Optional[str] = None
    description: Optional[str] = None
    default_options: Optional[Dict[str, Any]] = None


class RequestCommon(BaseModel, abc.ABC):
    """
    These are all possible values in a request from Active Workflow using the Remote Agent API.
    """

    method: str = Field(..., example="receive")
    params: ParamsCommon = Field(..., example=ParamsCommon())

    @validator("method")
    def method_choices(cls, v):
        if v not in ["register", "receive", "check"]:
            raise ValueError(
                "Method has to be one of 'register', 'receive' or 'check'."
            )
        return v


class ResponseCommon(BaseModel, abc.ABC):
    """
    This is a possible reponse returned when using the Acitve Workflow Remote Agent API.
    """

    result: ResultCommon = Field(..., example=ResultCommon())


#
# Application specifics
#


class PayloadInput(BaseModel):
    """
    This model represents the expected input payload when a message is received via `receive`.
    """

    context: Dict[str, Any] = Field({}, example={"test_val": 8})

    @staticmethod
    def example() -> Dict[str, Any]:
        return {
            "context": {"text_val": 8},
        }


class MessageOutput(BaseModel):
    """
    This model represents the produced output schema that will be provided in the messages.
    """

    job_run_id: int = Field(..., example=4)
    job_id: int = Field(..., example=1)
    job_name: str = Field(None, example="DummyJob")
    job_version: str = Field(None, example="1.0")
    completed: bool = Field(False, example=False)
    returncode: int = Field(1, example=1)
    runtime: float = Field(0.0, example=0.0)
    start_time: Optional[datetime.datetime] = Field(
        None, example=datetime.datetime(2020, 8, 14, 12, 2, 1)
    )
    end_time: Optional[datetime.datetime] = Field(
        None, example=datetime.datetime(2020, 8, 14, 12, 7, 29)
    )
    job_parameters: Dict[str, Any] = Field({}, example={"test_val": 8})
    response: Union[Dict[Any, Any], List[Any]] = Field(
        {}, example={"some": "returnval"}
    )

    @staticmethod
    def example() -> dict:
        return {
            "job_run_id": 4,
            "job_id": 1,
            "job_name": "DummyJob",
            "job_version": "1.0",
            "completed": False,
            "returncode": 1,
            "runtime": 0.0,
            "start_time": datetime.datetime(2020, 8, 14, 12, 2, 1).isoformat(),
            "end_time": datetime.datetime(2020, 8, 14, 12, 7, 29).isoformat(),
            "job_parameters": {},
            "response": {"some": "returnval"},
        }

    class Config:
        json_encoders = {
            datetime: lambda dt: dt.isoformat(),
        }


class OptionsCommon(BaseModel):
    """
    This model represents expected options to configure this agent (or rather a call to this agent).
    """

    job_name: str = Field(..., example="DummyJob")
    job_version: str = Field(..., example="1.0")
    talend_output_context_param: str = Field("", example="")
    api_key: str = Field("", example="")
    context: Dict[str, Any] = Field({}, example={"test_val": 8})

    @staticmethod
    def example() -> Dict[str, Any]:
        return {
            "job_name": "DummyJob",
            "job_version": "1.0",
            "talend_output_context_param": "",
            "api_key": "",
            "context": {"test_val": 8},
        }


class MemoryCommon(BaseModel):
    """
    This model represents the expected memory content to communicate state of the agent.
    """

    jobs: List[int] = Field([], example=[4, 5])

    @staticmethod
    def example() -> dict:
        return {"jobs": [4, 5]}


#
# Register Models
#


class ParamsRegister(ParamsCommon):
    pass


class RequestRegister(RequestCommon):
    method: str = Field("register", const="register")
    params: ParamsRegister = Field(ParamsRegister(), const=ParamsRegister())


class ResultRegister(ResultCommon):
    errors: Optional[List[str]] = None
    logs: Optional[List[str]] = None
    memory: Optional[Dict[str, Any]] = None
    messages: Optional[List[Dict[str, Any]]] = None
    name: str = config.BasicSettings().application_name.replace(" ", "")
    display_name: str = config.BasicSettings().application_name
    description: str = config.agent_description
    default_options: Dict[str, Any] = Field(
        OptionsCommon(**OptionsCommon.example()), example=OptionsCommon.example()
    )


class ResponseRegister(ResponseCommon):
    result: ResultRegister = ResultRegister()


#
# Receive Models
#


class MessageInputReceive(BaseModel):
    payload: PayloadInput


class ParamsReceive(ParamsCommon):
    message: MessageInputReceive = Field(
        ..., example=MessageInputReceive(payload=PayloadInput())
    )
    options: OptionsCommon = Field(
        OptionsCommon(**OptionsCommon.example()),
        example=OptionsCommon(**OptionsCommon.example()),
    )
    memory: MemoryCommon = Field(
        MemoryCommon(), example=MemoryCommon(**MemoryCommon.example())
    )
    credentials: List[CredentialsCommon] = Field([], example=[])


class RequestReceive(RequestCommon):
    method: str = Field("receive", const="receive")
    params: ParamsReceive = Field(
        ParamsReceive(
            message=MessageInputReceive(payload=PayloadInput(**PayloadInput.example()))
        ),
        example=ParamsReceive(message=MessageInputReceive(payload=PayloadInput())),
    )


class ResultReceive(ResultCommon):
    errors: List[str] = Field([], example=[])
    logs: List[str] = Field([], example=[])
    memory: MemoryCommon = Field(
        MemoryCommon(), example=MemoryCommon(**MemoryCommon.example())
    )
    messages: List[MessageOutput] = Field(
        [], example=[MessageOutput(**MessageOutput.example())]
    )


class ResponseReceive(ResponseCommon):
    result: ResultReceive = Field(ResultReceive(), example=ResultReceive())


#
# Check models
#


class ParamsCheck(ParamsCommon):
    message: Any = Field(None, const=None)
    options: OptionsCommon = Field(
        OptionsCommon(**OptionsCommon.example()),
        example=OptionsCommon(**OptionsCommon.example()),
    )
    memory: MemoryCommon = Field(
        MemoryCommon(), example=MemoryCommon(**MemoryCommon.example())
    )
    credentials: List[CredentialsCommon] = Field([])


class RequestCheck(RequestCommon):
    method: str = Field("check", const="check")
    params: ParamsCheck = Field(ParamsCheck(), example=ParamsCheck())


class ResultCheck(ResultCommon):
    errors: List[str] = Field([], example=[])
    logs: List[str] = Field([], example=[])
    memory: MemoryCommon = Field(
        MemoryCommon(), example=MemoryCommon(**MemoryCommon.example())
    )
    messages: List[MessageOutput] = Field(
        [], example=[MessageOutput(**MessageOutput.example())]
    )
    name: Optional[str] = None
    display_name: Optional[str] = None
    description: Optional[str] = None
    default_options: Optional[Dict[str, Any]] = None


class ResponseCheck(ResponseCommon):
    result: ResultCheck = Field(ResultCheck(), example=ResultCheck())
